import logging
from dataclasses import dataclass
from typing import cast, Any, Optional, Union, NamedTuple

import hydra
import numpy as np
import pandas as pd
import torch
from hydra.utils import instantiate
from omegaconf import DictConfig
from pydantic import BaseModel
from tokenizers import Encoding
from torch import optim
from torch.utils.data import DataLoader, Dataset
from transformers import (
    BatchEncoding,
    PreTrainedTokenizerBase,
)
from transformers.file_utils import PaddingStrategy

try:
    from pytorch_metric_learning import miners, losses
    from pytorch_lightning import Trainer, LightningModule
    from pytorch_lightning.utilities.types import (
        STEP_OUTPUT,
        TRAIN_DATALOADERS,
        EVAL_DATALOADERS,
        EPOCH_OUTPUT,
    )
except ImportError as e:
    raise ImportError(
        "Running the SapBERT model training code requires several additional dependencies to be installed.\n"
        "We recommend running 'pip install kazu[model-training]' to get all model training"
        " dependencies."
    ) from e

from kazu.utils.constants import HYDRA_VERSION_BASE
from kazu.utils.sapbert import SapBertHelper

logger = logging.getLogger(__name__)


@dataclass
class SapbertDataCollatorWithPadding:
    """Data collator to be used with HFSapbertPairwiseDataset."""

    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None

    def __call__(
        self, features: list[dict[str, BatchEncoding]]
    ) -> tuple[BatchEncoding, BatchEncoding]:
        query_toks1 = [x["query_toks1"] for x in features]
        query_toks1_enc = self.tokenizer.pad(
            query_toks1,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        query_toks2 = [x["query_toks2"] for x in features]
        query_toks2_enc = self.tokenizer.pad(
            query_toks2,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        return query_toks1_enc, query_toks2_enc


class HFSapbertPairwiseDataset(Dataset):
    """Dataset used for training SapBert."""

    def __getitem__(self, index: int) -> dict[str, Any]:
        query_toks1 = {
            "input_ids": self.encodings_1.data["input_ids"][index],
            "token_type_ids": self.encodings_1.data["token_type_ids"][index],
            "labels": self.encodings_1.data["labels"][index],
            "attention_mask": self.encodings_1.data["attention_mask"][index],
        }
        query_toks2 = {
            "input_ids": self.encodings_2.data["input_ids"][index],
            "token_type_ids": self.encodings_2.data["token_type_ids"][index],
            "labels": self.encodings_2.data["labels"][index],
            "attention_mask": self.encodings_2.data["attention_mask"][index],
        }

        return {"query_toks1": query_toks1, "query_toks2": query_toks2}

    def __init__(self, encodings_1: BatchEncoding, encodings_2: BatchEncoding, labels: np.ndarray):
        """
        :param encodings_1: encodings for example 1
        :param encodings_2: encodings for example 2
        :param labels: labels i.e. knowledgebase identifier for both encodings, as an int
        """
        encodings_1["labels"] = labels
        encodings_2["labels"] = labels
        self.encodings_1 = encodings_1
        self.encodings_2 = encodings_2

    def __len__(self):
        encodings = cast(list[Encoding], self.encodings_1.encodings)
        return len(encodings)


class SapbertTrainingParams(BaseModel):
    topk: int  # score this many nearest neighbours against target
    lr: float  # learning rate
    weight_decay: float
    miner_margin: float  # passed to TripletMarginMiner
    type_of_triplets: str  # passed to TripletMarginMiner
    train_file: str  # a parquet file with three columns - syn1, syn2 and id
    train_batch_size: int
    num_workers: int  # passed to dataloaders


class SapbertEvaluationDataset(NamedTuple):
    """To evaluate a given embedding model, we need a query datasource (i.e. things that
    need to be linked)] and an ontology datasource (i.e. things that we need to generate
    an embedding space for, that can be queried against) each should have three columns:

    default_label (text), iri (ontology id) and source (ontology name)
    """

    query_source: pd.DataFrame
    ontology_source: pd.DataFrame


class Candidate(NamedTuple):
    """A knowledgebase entry."""

    default_label: str
    iri: str
    correct: bool


class GoldStandardExample(NamedTuple):
    gold_default_label: str
    gold_iri: str
    candidates: list[
        Candidate
    ]  # candidates (aka nearest neighbours) associate with this gold instance


class SapbertEvaluationDataManager:
    """Manages the loading/parsing of multiple evaluation datasets. Each dataset should
    have two sources, a query source and an ontology source. these are then converted
    into data loaders, while maintaining a reference to the embedding metadata that
    should be used for evaluation.

    self.dataset is dict[dataset_name, SapbertEvaluationDataset] after construction
    """

    def __init__(self, sources: dict[str, list[str]], debug: bool = False):
        self.datasets: dict[str, SapbertEvaluationDataset] = {}
        for source_name, (
            query_source_path,
            ontology_source_path,
        ) in sources.items():
            query_source = pd.read_parquet(query_source_path)
            ontology_source = pd.read_parquet(ontology_source_path)
            if debug:
                query_source = query_source.head(10)
                ontology_source = ontology_source.sample(frac=1.0).head(100)
            self.datasets[source_name] = SapbertEvaluationDataset(
                query_source=query_source, ontology_source=ontology_source
            )


class PLSapbertModel(LightningModule):
    def __init__(
        self,
        model_name_or_path: str,
        sapbert_training_params: Optional[SapbertTrainingParams] = None,
        sapbert_evaluation_manager: Optional[SapbertEvaluationDataManager] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        :param model_name_or_path: passed to AutoModel.from_pretrained
        :param sapbert_training_params: optional SapbertTrainingParams, only needed if training a model
        :param sapbert_evaluation_manager: optional SapbertEvaluationDataManager, only needed if training a model
        :param args: passed to LightningModule
        :param kwargs: passed to LightningModule
        """

        super().__init__(*args, **kwargs)
        self.sapbert_helper = SapBertHelper(model_name_or_path)
        self.model = self.sapbert_helper.model
        self.sapbert_evaluation_manager = sapbert_evaluation_manager
        self.sapbert_training_params = sapbert_training_params
        if sapbert_training_params is not None:
            self.loss = losses.MultiSimilarityLoss(alpha=1, beta=60, base=0.5)
            self.miner = miners.TripletMarginMiner(
                margin=sapbert_training_params.miner_margin,
                type_of_triplets=sapbert_training_params.type_of_triplets,
            )

    def configure_optimizers(self):
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.configure_optimizers
        </common/lightning_module.rst#configure-optimizers>`\\ ."""
        assert self.sapbert_training_params is not None
        optimizer = optim.AdamW(
            [
                {"params": self.model.parameters()},
            ],
            lr=self.sapbert_training_params.lr,
            weight_decay=self.sapbert_training_params.weight_decay,
        )
        return optimizer

    def forward(self, batch: BatchEncoding) -> dict[int, torch.Tensor]:
        """For inference.

        :param batch: standard bert input, with an additional 'indices' for representing
            the location of the embedding
        :return:
        """
        return self.sapbert_helper.get_prediction_from_batch(self.model, batch)

    def training_step(self, batch: Any, batch_idx: int, *args: Any, **kwargs: Any) -> STEP_OUTPUT:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.training_step
        </common/lightning_module.rst#training-step>`\\ ."""
        query_toks1, query_toks2 = batch
        # labels should be identical, so we only need one
        labels = query_toks1.pop("labels")
        query_toks2.pop("labels")
        last_hidden_state1 = self.model(**query_toks1, return_dict=True).last_hidden_state
        last_hidden_state2 = self.model(**query_toks2, return_dict=True).last_hidden_state
        query_embed1 = last_hidden_state1[:, 0]  # query : [batch_size, hidden]
        query_embed2 = last_hidden_state2[:, 0]  # query : [batch_size, hidden]
        query_embed = torch.cat([query_embed1, query_embed2], dim=0)
        labels = torch.cat([labels, labels], dim=0)
        hard_pairs = self.miner(query_embed, labels)
        return self.loss(query_embed, labels, hard_pairs)  # type: ignore [no-any-return] # no clear type info for the loss

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.train_dataloader
        </common/lightning_module.rst#train-dataloader>`\\ ."""
        assert self.sapbert_training_params is not None
        training_df = pd.read_parquet(self.sapbert_training_params.train_file)
        labels = training_df["id"].astype("category").cat.codes.to_numpy()
        encodings_1 = self.tokeniser(training_df["syn1"].tolist())
        encodings_2 = self.tokeniser(training_df["syn2"].tolist())
        encodings_1["labels"] = labels
        encodings_2["labels"] = labels
        train_set = HFSapbertPairwiseDataset(
            labels=labels, encodings_1=encodings_1, encodings_2=encodings_2
        )
        train_loader = DataLoader(
            train_set,
            batch_size=self.sapbert_training_params.train_batch_size,
            shuffle=True,
            num_workers=self.sapbert_training_params.num_workers,
            collate_fn=SapbertDataCollatorWithPadding(
                self.tokeniser, padding=PaddingStrategy.LONGEST
            ),
        )

        return train_loader

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.val_dataloader
        </common/lightning_module.rst#val-dataloader>`\\ ."""
        dataloaders = []
        assert self.sapbert_evaluation_manager is not None
        assert self.sapbert_training_params is not None
        for query_source, ontology_source in self.sapbert_evaluation_manager.datasets.values():
            query_dataloader = self.sapbert_helper.get_embedding_dataloader_from_strings(
                texts=query_source["default_label"].tolist(),
                batch_size=self.sapbert_training_params.train_batch_size,
                num_workers=self.sapbert_training_params.num_workers,
            )
            ontology_dataloader = self.sapbert_helper.get_embedding_dataloader_from_strings(
                texts=ontology_source["default_label"].tolist(),
                batch_size=self.sapbert_training_params.train_batch_size,
                num_workers=self.sapbert_training_params.num_workers,
            )
            dataloaders.append(query_dataloader)
            dataloaders.append(ontology_dataloader)

        return dataloaders

    def validation_step(
        self, batch: Any, batch_idx: int, dataset_idx: int
    ) -> Optional[STEP_OUTPUT]:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.validation_step
        </common/lightning_module.rst#validation-step>`\\ ."""
        return self(batch)  # type: ignore [no-any-return] # no type info from Pytorch Lightning

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: Optional[int] = None) -> Any:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.predict_step
        </common/lightning_module.rst#predict-step>`\\ ."""
        return self(batch)

    def validation_epoch_end(self, outputs: Union[EPOCH_OUTPUT, list[EPOCH_OUTPUT]]) -> None:
        """Lightning override generate new embeddings for each
        :attr:`SapbertEvaluationDataset.ontology_source` and query them with
        :attr:`SapbertEvaluationDataset.query_source`

        :param outputs:
        :return:
        """
        assert self.sapbert_evaluation_manager is not None
        raise NotImplementedError

    def log_results(self, dataset_name, metrics):
        for key, val in metrics.items():
            if key.startswith("acc"):
                self.log(key, value=val, rank_zero_only=True)
                logger.info(f"{dataset_name}: {key}, {val}")

    def get_candidate_dict(self, np_candidates: pd.DataFrame, golden_iri: str) -> list[Candidate]:
        """Convert rows in a dataframe representing candidate KB entries into a
        corresponding :class:`Candidate` per row.

        :param np_candidates:
        :param golden_iri:
        :return:
        """
        candidates_filtered = []
        for i, np_candidate_row in np_candidates.iterrows():
            candidates_filtered.append(
                Candidate(
                    default_label=np_candidate_row["default_label"],
                    iri=np_candidate_row["iri"],
                    correct=np_candidate_row["iri"] == golden_iri,
                )
            )
        return candidates_filtered

    def evaluate_topk_acc(self, queries: list[GoldStandardExample]) -> dict[str, float]:
        """Get a dictionary of accuracy results at different levels of k (nearest
        neighbours)

        :param queries:
        :return:
        """
        k = len(queries[0].candidates)
        result = {}
        for i in range(0, k):
            hit = 0
            for query in queries:
                candidates = query.candidates[: i + 1]  # to get acc@(i+1)
                if any(candidate.correct for candidate in candidates):
                    hit += 1
            result["acc{}".format(i + 1)] = hit / len(queries)

        return result


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path="../../../conf", config_name="config")
def start(cfg: DictConfig) -> None:
    trainer: Trainer = instantiate(cfg.SapBertTraining.trainer)
    model: PLSapbertModel = instantiate(cfg.SapBertTraining.model)
    trainer.fit(model)


if __name__ == "__main__":
    start()



"""Some functions to update public resources in the kazu model pack."""
import argparse
import shutil
import sqlite3
import subprocess
from pathlib import Path

import requests

_ONTOLOGY_DIR_NAME = "ontologies"

ONTOLOGY_NAMES_TO_PATH = {
    "CL": (
        "http://purl.obolibrary.org/obo/cl.owl",
        "cl.owl",
    ),
    "CLO": (
        "http://purl.obolibrary.org/obo/clo.owl",
        "clo.owl",
    ),
    "GO": (
        "http://purl.obolibrary.org/obo/go.owl",
        "go.owl",
    ),
    "MONDO": (
        "http://purl.obolibrary.org/obo/mondo.json",
        "mondo.json",
    ),
    "STATO": (
        "http://purl.obolibrary.org/obo/stato.owl",
        "stato.owl",
    ),
    "UBERON": (
        "http://purl.obolibrary.org/obo/uberon.owl",
        "uberon.owl",
    ),
    "HGNC_GENE_FAMILY": (
        "http://biomart.genenames.org/martservice/results?query="
        '<!DOCTYPE Query><Query client="biomartclient" processor="TSV" limit="-1" header="1">'
        '<Dataset name="hgnc_family_mart" config="family_config">'
        '<Attribute name="family__family_id_103"/><Attribute name="family__name_103"/>'
        '<Attribute name="family__alias__alias_101"/><Attribute name="family__root_symbol_103"/><'
        "/Dataset></Query>",
        "gene_family.tsv",
    ),
    "CELLOSAURUS": (
        "https://ftp.expasy.org/databases/cellosaurus/cellosaurus.obo",
        "cellosaurus.obo",
    ),
    "HPO": (
        "https://github.com/obophenotype/human-phenotype-ontology/releases/latest/download/hp-full.owl",
        "hp-full.owl",
    ),
}

DOWNLOAD_KEYS = list(ONTOLOGY_NAMES_TO_PATH.keys()) + ["OPENTARGETS", "CHEMBL"]


def download_and_process_chembl(output_dir: Path, chembl_version: str) -> None:
    target_dir = output_dir.joinpath(_ONTOLOGY_DIR_NAME).absolute()
    chembl_url = f"https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_{chembl_version}/chembl_{chembl_version}_sqlite.tar.gz"

    print("downloading chembl")
    subprocess.run(
        ["wget", chembl_url],
        cwd=target_dir,
    )
    print("extracting chembl DB")
    subprocess.run(["tar", "-xvzf", chembl_url.split("/")[-1]], cwd=target_dir)
    chembl_db_path = (
        target_dir.joinpath(f"chembl_{chembl_version}")
        .joinpath(f"chembl_{chembl_version}_sqlite")
        .joinpath(f"chembl_{chembl_version}.db")
    )

    conn = sqlite3.connect(chembl_db_path)
    LIST_TABLES = """SELECT
        name
    FROM
        sqlite_schema
    WHERE
        type ='table' AND
        name NOT LIKE 'sqlite_%';"""
    cur = conn.cursor()
    tables = cur.execute(LIST_TABLES)
    print("removing superfluous chembl tables to save space")
    for table_tup in tables.fetchall():
        table_name = table_tup[0]
        if table_name not in {
            "molecule_atc_classification",
            "molecule_dictionary",
            "molecule_hierarchy",
            "molecule_synonyms",
        }:
            print(f"dropping table {table_name}")
            conn.execute(f"DROP TABLE {table_name}")
    cur.close()
    print("running sqllite VACUUM")
    conn = sqlite3.connect(chembl_db_path, isolation_level=None)
    conn.execute("VACUUM")
    conn.close()


def download_single_file_resources(output_dir: Path, resources_to_download: list[str]) -> None:
    for resource_to_download in resources_to_download:
        maybe_url_path = ONTOLOGY_NAMES_TO_PATH.get(resource_to_download)
        if maybe_url_path is not None:
            url, local_path = maybe_url_path
            print(f"downloading data from {url}")
            response = requests.get(url, stream=True)
            if response.status_code > 300:
                print(f"request failed: {url}")
                print(response)
            else:
                with output_dir.joinpath(_ONTOLOGY_DIR_NAME).joinpath(local_path).open(
                    mode="wb"
                ) as f:
                    for chunk in response.iter_content(chunk_size=None):
                        f.write(chunk)


def download_ftp_resources(output_dir: Path, open_targets_version: str) -> None:
    path_to_ftp_directories = {
        "opentargets/": [
            f"ftp://ftp.ebi.ac.uk/pub/databases/opentargets/platform/{open_targets_version}/output/etl/json/targets"
            f"ftp://ftp.ebi.ac.uk/pub/databases/opentargets/platform/{open_targets_version}/output/etl/json/diseases"
            f"ftp://ftp.ebi.ac.uk/pub/databases/opentargets/platform/{open_targets_version}/output/etl/json/molecule"
        ]
    }
    for path_str, ftp_directories in path_to_ftp_directories.items():
        shutil.rmtree(path_str, ignore_errors=True)
        output_path = output_dir.joinpath(_ONTOLOGY_DIR_NAME).joinpath(path_str)
        output_path.mkdir(parents=True, exist_ok=True)
        for url in ftp_directories:

            subprocess.run(
                [
                    "wget",
                    "--recursive",
                    "--no-parent",
                    "--no-host-directories",
                    "--cut-dirs",
                    "8",
                    url,
                ],
                cwd=output_path,
            )


if __name__ == "__main__":
    description = "Download public ontologies for Kazu."

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--open_targets_release_version",
        type=str,
        required=False,
        help="Version of OpenTargets to use, e.g. '23.09'",
    )
    parser.add_argument(
        "--chembl_version",
        type=str,
        required=False,
        help="Version of Chembl to use, e.g. '33'",
    )
    parser.add_argument(
        "--output_dir",
        type=Path,
        required=True,
        help="path to download resources to",
    )
    parser.add_argument(
        "--resources_to_download",
        nargs="+",
        required=False,
        default=DOWNLOAD_KEYS,
        help=f"resources to download. Choose from {DOWNLOAD_KEYS}. If blank, download all",
    )
    args = parser.parse_args()

    output_dir: Path = args.output_dir
    output_dir.joinpath(_ONTOLOGY_DIR_NAME).mkdir(parents=True, exist_ok=True)
    download_single_file_resources(output_dir, args.resources_to_download)

    if "OPENTARGETS" in args.resources_to_download:
        if args.open_targets_release_version:
            download_ftp_resources(output_dir, args.open_targets_release_version)
        else:
            print(
                "skipping open targets as no version was provided with --open_targets_release_version"
            )
    if "CHEMBL" in args.resources_to_download:
        if args.chembl_version:
            download_and_process_chembl(output_dir, args.chembl_version)
        else:
            print("skipping chembl as no version was provided with --chembl_version")



import dataclasses
import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict, Counter
from enum import auto
from typing import cast, Optional, Literal, Any
from collections.abc import Iterable

import pandas as pd
from kazu.data.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    SynonymTerm,
    SimpleValue,
    CuratedTerm,
    ParserBehaviour,
    CuratedTermBehaviour,
    AssociatedIdSets,
    GlobalParserActions,
    AutoNameEnum,
    MentionConfidence,
    IdsAndSource,
)
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
    NormalisedSynonymStr,
    Idx,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.synonym_generation import CombinatorialSynonymGenerator
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.string_normalizer import StringNormalizer
from kazu.utils.utils import (
    PathLike,
    as_path,
    string_to_putative_curation,
    extract_term_strings_from_synonym_terms,
)

#: The column name in a dataframe parsed with :meth:`~.OntologyParser.parse_to_dataframe`
#: for the column of the entity's default/preferred label
DEFAULT_LABEL = "default_label"
#: The column name for the id of each entity
IDX = "idx"
#: The column name for the synonyms/alternative labels for each entity
SYN = "syn"
#: The column name for the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology
MAPPING_TYPE = "mapping_type"
#: The origin of a dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc.
#: Note, this is different from the parser.name, as is used to identify the origin of a mapping back to a data source
DATA_ORIGIN = "data_origin"

logger = logging.getLogger(__name__)


class CurationException(Exception):
    pass


def load_curated_terms(
    path: PathLike,
) -> list[CuratedTerm]:
    """Load :class:`kazu.data.data.CuratedTerm`\\ s from a file path.

    :param path: path to json lines file that map to :class:`kazu.data.data.CuratedTerm`
    :return:
    """
    curations_path = as_path(path)
    if curations_path.exists():
        with curations_path.open(mode="r") as jsonlf:
            curations = [CuratedTerm.from_json(line) for line in jsonlf]
    else:
        raise ValueError(f"curations do not exist at {path}")
    return curations


def load_global_actions(
    path: PathLike,
) -> GlobalParserActions:
    """Load an instance of :class:`.GlobalParserActions` from a file path.

    :param path: path to a json serialised GlobalParserActions
    :return:
    """
    global_actions_path = as_path(path)
    if global_actions_path.exists():
        with global_actions_path.open(mode="r") as jsonf:
            global_actions = GlobalParserActions.from_json(json.load(jsonf))
    else:
        raise ValueError(f"global actions do not exist at {path}")
    return global_actions


class CurationModificationResult(AutoNameEnum):
    ID_SET_MODIFIED = auto()
    SYNONYM_TERM_ADDED = auto()
    SYNONYM_TERM_DROPPED = auto()
    NO_ACTION = auto()


class CurationProcessor:
    """A CurationProcessor is responsible for modifying the set of
    :class:`.SynonymTerm`\\s produced by an
    :class:`kazu.ontology_preprocessing.base.OntologyParser` with any relevant
    :class:`.GlobalParserActions` and/or :class:`.CuratedTerm` associated with the
    parser.

    That is to say, this class modifies the raw data produced by a parser with any a posteriori
    observations about the data (such as conflicts in case sensitivity/identifiers/mention confidence etc). Is also
    identifies curations that should be used for dictionary based NER.

    Note that this class does not handle inherited behaviour. For instance, if a :class:`.CuratedTerm` has the behaviour
    INHERIT_FROM_SOURCE_TERM, and the source terms behaviour is DROP_SYNONYM_TERM_FOR_LINKING, it is the responsibility
    of the caller to ensure that the inherited curation is handled correctly.
    :func:`kazu.utils.curated_term_tools.filter_curations_for_ner` may be useful in this case.

    This class should be used before instances of :class:`.SynonymTerm`\\s are loaded into the
    internal database representation.
    """

    # curations are applied in the following order
    CURATION_APPLY_ORDER = (
        CuratedTermBehaviour.IGNORE,
        CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
        CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
        CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING,
        CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM,
    )
    _BEHAVIOUR_TO_ORDER_INDEX = {behav: i for i, behav in enumerate(CURATION_APPLY_ORDER)}

    def __init__(
        self,
        parser_name: str,
        entity_class: str,
        global_actions: Optional[GlobalParserActions],
        curations: list[CuratedTerm],
        synonym_terms: set[SynonymTerm],
    ):
        """

        :param parser_name: name of parser to process
        :param entity_class: name of parser entity_class to process (typically as passed to :class:`kazu.ontology_preprocessing.base.OntologyParser`\\ )
        :param global_actions:
        :param curations:
        :param synonym_terms:
        """
        self.global_actions = global_actions
        self.entity_class = entity_class
        self.parser_name = parser_name
        self._terms_by_term_norm: dict[NormalisedSynonymStr, SynonymTerm] = {}
        self._terms_by_id: defaultdict[Idx, set[SynonymTerm]] = defaultdict(set)
        # used by inherited curations to decide behaviour
        self._curations_by_syn: defaultdict[str, set[CuratedTerm]] = defaultdict(set)
        for term in synonym_terms:
            self._update_term_lookups(term, False)
        self.curations = set(curations)
        self.dropped_keys: set[NormalisedSynonymStr] = set()

    @classmethod
    def curation_sort_key(cls, curated_term: CuratedTerm) -> tuple[int, bool, str]:
        """Determines the order curations are processed in.

        We use associated_id_sets as a key, so that any overrides will be processed
        after any original behaviours
        """
        return (
            cls._BEHAVIOUR_TO_ORDER_INDEX[curated_term.behaviour],
            curated_term.associated_id_sets is not None,
            curated_term.curated_synonym,
        )

    def _update_term_lookups(
        self, term: SynonymTerm, override: bool
    ) -> Literal[
        CurationModificationResult.SYNONYM_TERM_ADDED, CurationModificationResult.NO_ACTION
    ]:
        assert term.original_term is None

        safe_to_add = False
        maybe_existing_term = self._terms_by_term_norm.get(term.term_norm)
        if maybe_existing_term is None:
            logger.debug("adding new term %s", term)
            safe_to_add = True
        elif override:
            safe_to_add = True
            logger.debug("overriding existing term %s", maybe_existing_term)
        elif (
            len(
                term.associated_id_sets.symmetric_difference(maybe_existing_term.associated_id_sets)
            )
            > 0
        ):
            logger.warning(
                "conflict on term norms \n%s\n%s\nthe latter will be ignored",
                maybe_existing_term,
                term,
            )
        if safe_to_add:
            self._terms_by_term_norm[term.term_norm] = term
            for equiv_ids in term.associated_id_sets:
                for idx in equiv_ids.ids:
                    self._terms_by_id[idx].add(term)
            return CurationModificationResult.SYNONYM_TERM_ADDED
        else:
            return CurationModificationResult.NO_ACTION

    def _drop_synonym_term(self, synonym: NormalisedSynonymStr) -> None:
        """Remove a synonym term from the database, so that it cannot be used as a
        linking target.

        :param synonym:
        :return:
        """
        try:
            term_to_remove = self._terms_by_term_norm.pop(synonym)
            self.dropped_keys.add(synonym)
            for equiv_id_set in term_to_remove.associated_id_sets:
                for idx in equiv_id_set.ids:
                    terms_by_id = self._terms_by_id.get(idx)
                    if terms_by_id is not None:
                        terms_by_id.remove(term_to_remove)
            logger.debug(
                "successfully dropped %s from database for %s",
                synonym,
                self.entity_class,
            )
        except KeyError:
            if synonym in self.dropped_keys:
                logger.debug(
                    "tried to drop %s from database, but key already dropped by another CuratedTerm for %s",
                    synonym,
                    self.parser_name,
                )
            else:
                logger.warning(
                    "tried to drop %s from database, but key doesn't exist for %s",
                    synonym,
                    self.parser_name,
                )

    def _drop_id_from_all_synonym_terms(self, id_to_drop: Idx) -> Counter:
        """Remove a given id from all :class:`.SynonymTerm`\\ s.

        Drop any :class:`.SynonymTerm`\\ s with no remaining ID after removal.

        :param id_to_drop:
        :return: counter of :class:`.CurationModificationResult`
        """

        terms_to_modify = self._terms_by_id.get(id_to_drop, set())
        counter = Counter(
            self._drop_id_from_synonym_term(id_to_drop=id_to_drop, term_to_modify=term_to_modify)
            for term_to_modify in set(terms_to_modify)
        )

        return counter

    def _drop_id_from_synonym_term(
        self, id_to_drop: Idx, term_to_modify: SynonymTerm
    ) -> Literal[
        CurationModificationResult.ID_SET_MODIFIED,
        CurationModificationResult.SYNONYM_TERM_DROPPED,
        CurationModificationResult.NO_ACTION,
    ]:
        """Remove an id from a given :class:`.SynonymTerm`\\ .

        :param id_to_drop:
        :param term_to_modify:
        :return:
        """
        new_assoc_id_frozenset = self._drop_id_from_associated_id_sets(
            id_to_drop, term_to_modify.associated_id_sets
        )
        if len(new_assoc_id_frozenset.symmetric_difference(term_to_modify.associated_id_sets)) == 0:
            return CurationModificationResult.NO_ACTION
        else:
            return self._modify_or_drop_synonym_term_after_id_set_change(
                new_associated_id_sets=new_assoc_id_frozenset, synonym_term=term_to_modify
            )

    def _drop_id_from_associated_id_sets(
        self, id_to_drop: Idx, associated_id_sets: AssociatedIdSets
    ) -> AssociatedIdSets:
        """Remove an id from a :class:`.AssociatedIdSets`\\ .

        :param id_to_drop:
        :param associated_id_sets:
        :return:
        """
        new_assoc_id_set = set()
        for equiv_id_set in associated_id_sets:
            if id_to_drop in equiv_id_set.ids:
                updated_ids_and_source = frozenset(
                    id_tup for id_tup in equiv_id_set.ids_and_source if id_tup[0] != id_to_drop
                )
                if len(updated_ids_and_source) > 0:
                    updated_equiv_id_set = EquivalentIdSet(updated_ids_and_source)
                    new_assoc_id_set.add(updated_equiv_id_set)
            else:
                new_assoc_id_set.add(equiv_id_set)
        new_assoc_id_frozenset = frozenset(new_assoc_id_set)
        return new_assoc_id_frozenset

    def _modify_or_drop_synonym_term_after_id_set_change(
        self, new_associated_id_sets: AssociatedIdSets, synonym_term: SynonymTerm
    ) -> Literal[
        CurationModificationResult.ID_SET_MODIFIED, CurationModificationResult.SYNONYM_TERM_DROPPED
    ]:
        """Modifies or drops a :class:`.SynonymTerm` after a :class:`.AssociatedIdSets`
        has changed.

        :param new_associated_id_sets:
        :param synonym_term:
        :return:
        """
        result: Literal[
            CurationModificationResult.ID_SET_MODIFIED,
            CurationModificationResult.SYNONYM_TERM_DROPPED,
        ]
        if len(new_associated_id_sets) > 0:
            if new_associated_id_sets == synonym_term.associated_id_sets:
                raise ValueError(
                    "function called inappropriately where the id sets haven't changed. This"
                    " has failed as it will otherwise modify the value of aggregated_by, when"
                    " nothing has changed"
                )
            new_term = dataclasses.replace(
                synonym_term,
                associated_id_sets=new_associated_id_sets,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            add_result = self._update_term_lookups(new_term, True)
            assert add_result is CurationModificationResult.SYNONYM_TERM_ADDED
            result = CurationModificationResult.ID_SET_MODIFIED
        else:
            # if there are no longer any id sets associated with the record, remove it completely
            self._drop_synonym_term(synonym_term.term_norm)
            result = CurationModificationResult.SYNONYM_TERM_DROPPED
        return result

    def export_curations_and_final_terms(
        self,
    ) -> tuple[list[CuratedTerm], set[SynonymTerm]]:
        """Perform any updates required to the synonym terms as specified in the
        curations/global actions.

        The returned :class:`.CuratedTerm`\\s can be used for Dictionary based NER, whereas the
        returned :class:`.SynonymTerm`\\s can be loaded into the internal database for linking.

        :return:
        """
        self._process_global_actions()
        return list(self._process_curations()), set(self._terms_by_term_norm.values())

    def _process_curations(self) -> Iterable[CuratedTerm]:
        safe_curations = self.fix_conflicts_in_curations(self.curations)
        for curation in sorted(safe_curations, key=self.curation_sort_key):
            curation = self._process_curation_action(curation)
            yield curation

    def fix_conflicts_in_curations(self, curations: set[CuratedTerm]) -> set[CuratedTerm]:
        """Check to see if a list of curations contain conflicts.

        Conflicts can occur for the following reasons:

        1) If two or more curations normalise to the same NormalisedSynonymStr,
           but have different behaviours.

        2) If two or more curations normalise to the same NormalisedSynonymStr,
           but have different associated ID sets specified, such that one would
           override the other.

        3) If two or more curations have conflicting values for case sensitivity and
           mention_confidence. E.g. A case-insensitive curation cannot have a higher
           mention confidence value than a case-sensitive one for the same synonym.


        :param curations:
        :return: safe curations set
        """

        curations_by_term_norm = defaultdict(set)
        curations_by_syn_lower = defaultdict(set)
        for curation in curations:
            curations_by_term_norm[curation.term_norm_for_linking(self.entity_class)].add(curation)
            curations_by_syn_lower[curation.curated_synonym.lower()].add(curation)

        all_remove = set()
        for potential_conflict_set in curations_by_term_norm.values():
            if len(potential_conflict_set) > 1:
                to_add, to_remove = self.resolve_behaviour_conflicts(potential_conflict_set)
                all_remove.update(to_remove)
                curations.update(to_add)

        for potential_conflict_set in curations_by_syn_lower.values():
            potential_conflict_set.difference_update(all_remove)
            if len(potential_conflict_set) > 1:
                to_add, to_remove = self.resolve_case_conflicts(potential_conflict_set)
                all_remove.update(to_remove)
                curations.update(to_add)

        curations.difference_update(all_remove)
        return curations

    def resolve_behaviour_conflicts(
        self, curations: set[CuratedTerm]
    ) -> tuple[set[CuratedTerm], set[CuratedTerm]]:

        curations_by_syn_lower = defaultdict(set)
        potentially_conflicting_behaviours = set()
        potentially_conflicting_id_sets = set()
        source_curations = set()
        inherited_curations = set()
        for curation in curations:
            if curation.source_term is None:
                source_curations.add(curation)
                if curation.behaviour in {
                    CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
                    CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
                    CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING,
                }:
                    potentially_conflicting_behaviours.add(curation.behaviour)
                if curation.associated_id_sets is not None:
                    potentially_conflicting_id_sets.add(curation.associated_id_sets)
            else:
                inherited_curations.add(curation)
            curations_by_syn_lower[curation.curated_synonym.lower()].add(curation)

        if len(potentially_conflicting_behaviours) > 1:
            resolved_behaviour = (
                CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING
                if CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING
                in potentially_conflicting_behaviours
                else CuratedTermBehaviour.IGNORE
            )
            logger.warning(
                "conflicting behaviours detected. The following source curations will be\n%s\n%s",
                resolved_behaviour.name,
                "\n\n".join(curation.to_json() for curation in source_curations) + "\n\n",
            )
            if len(inherited_curations) > 0:
                logger.warning(
                    "conflicting behaviours detected. The following inherited curations will be\n%s\n%s",
                    CuratedTermBehaviour.IGNORE.name,
                    "\n\n".join(curation.to_json() for curation in inherited_curations) + "\n\n",
                )
            return (
                set(
                    dataclasses.replace(curation, behaviour=resolved_behaviour)
                    for curation in source_curations
                ).union(
                    dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
                    for curation in inherited_curations
                ),
                curations,
            )
        if len(potentially_conflicting_id_sets) > 1:
            raise CurationException(
                "conflicting id sets detected in curations. Please fix the below curations\n%s",
                "\n\n".join(curation.to_json() for curation in source_curations)
                + "\n\n"
                + "\n\n".join(curation.to_json() for curation in inherited_curations)
                + "\n\n",
            )
        return set(), set()

    def resolve_case_conflicts(
        self, curations: set[CuratedTerm]
    ) -> tuple[set[CuratedTerm], set[CuratedTerm]]:

        to_add: set[CuratedTerm] = set()
        to_remove: set[CuratedTerm] = set()
        cs_conf = set()
        ci_conf = set()
        cs_lookup: defaultdict[str, tuple[set[CuratedTerm], set[MentionConfidence]]] = defaultdict(
            lambda: (set(), set())
        )
        for curation in curations:
            if curation.behaviour in {
                CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM,
                CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
            }:
                if curation.case_sensitive:
                    cs_conf.add(curation.mention_confidence)
                    cs_lookup[curation.curated_synonym][0].add(curation)
                    cs_lookup[curation.curated_synonym][1].add(curation.mention_confidence)
                else:
                    ci_conf.add(curation.mention_confidence)

        if len(ci_conf.union(cs_conf)) == 1:
            logger.debug("curations OK \n %s", curations)
        elif len(ci_conf) > 1 or (
            len(ci_conf) > 0 and len(cs_conf) > 0 and min(ci_conf) < min(cs_conf)
        ):
            # set all to CI min
            target_conf = min(ci_conf)
            logger.warning(
                "conflict detected in case sensitivity/confidence combination for the following curations. Affected curations will adopt the conservative confidence value\n%s\n%s",
                target_conf.name,
                "\n\n".join(curation.to_json() for curation in curations) + "\n\n",
            )
            to_add.update(
                dataclasses.replace(curation, mention_confidence=target_conf)
                for curation in curations
            )
            to_remove.update(curations)
        elif len(cs_conf) > 1:
            for curations_cs, cs_confs_by_exact_match in cs_lookup.values():
                if len(cs_confs_by_exact_match) > 1:
                    target_conf = min(cs_confs_by_exact_match)
                    logger.warning(
                        "conflict detected in case sensitivity/confidence combination for the following curations. Affected curations will adopt the conservative confidence value %s,%s",
                        target_conf.name,
                        "\n".join(curation.to_json() for curation in curations_cs) + "\n\n",
                    )
                    to_add.update(
                        dataclasses.replace(curation, mention_confidence=target_conf)
                        for curation in curations_cs
                    )
                    to_remove.update(curations_cs)

        return to_add, to_remove

    def _process_curation_action(self, curation: CuratedTerm) -> CuratedTerm:

        if curation.source_term is None:
            self._curations_by_syn[curation.curated_synonym].add(curation)
        term_norm = curation.term_norm_for_linking(self.entity_class)
        if curation.behaviour is CuratedTermBehaviour.IGNORE:
            logger.debug("curation ignored: %s for %s", curation, self.parser_name)
        elif curation.behaviour is CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM:
            logger.debug(
                "curation inherits behaviour from %s for %s",
                curation.source_term,
                self.parser_name,
            )
            return curation
        elif curation.behaviour is CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING:
            self._drop_synonym_term(term_norm)
        elif curation.behaviour is CuratedTermBehaviour.ADD_FOR_LINKING_ONLY:
            self._attempt_to_add_database_entry_for_curation(
                curation_associated_id_set=curation.associated_id_sets,
                curated_synonym=curation.curated_synonym,
                curation_term_norm=term_norm,
            )

        elif curation.behaviour is CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING:
            self._attempt_to_add_database_entry_for_curation(
                curation_associated_id_set=curation.associated_id_sets,
                curated_synonym=curation.curated_synonym,
                curation_term_norm=term_norm,
            )
            term_for_this_curation = self._terms_by_term_norm.get(term_norm)
            if term_for_this_curation is None:
                logger.warning(
                    "CuratedTerm %s is invalid: "
                    "requires an identifier but none was found. It may have been removed by another curation, or not exist in the underlying data sourcee.",
                    curation,
                )
                return dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
        else:
            raise ValueError(f"unknown behaviour for parser {self.parser_name}, {curation}")
        return curation

    def _process_global_actions(self) -> None:
        if self.global_actions is None:
            return None

        override_curations_by_id = defaultdict(set)
        for curation in self.curations:
            if curation.associated_id_sets is not None:
                for equiv_id_set in curation.associated_id_sets:
                    for idx in equiv_id_set.ids:
                        override_curations_by_id[idx].add(curation)

        for action in self.global_actions.parser_behaviour(self.parser_name):
            if action.behaviour is ParserBehaviour.DROP_IDS_FROM_PARSER:
                ids = action.parser_to_target_id_mappings[self.parser_name]
                for idx in ids:
                    counter_this_idx = self._drop_id_from_all_synonym_terms(idx)
                    if (
                        counter_this_idx[CurationModificationResult.ID_SET_MODIFIED]
                        + counter_this_idx[CurationModificationResult.SYNONYM_TERM_DROPPED]
                        == 0
                    ):
                        logger.warning("failed to drop %s from %s", idx, self.parser_name)
                    else:
                        logger.debug(
                            "dropped ID %s from %s. SynonymTerm modified count: %s, SynonymTerm dropped count: %s",
                            idx,
                            self.parser_name,
                            counter_this_idx[CurationModificationResult.ID_SET_MODIFIED],
                            counter_this_idx[CurationModificationResult.SYNONYM_TERM_DROPPED],
                        )

                        for override_curation_to_modify in set(
                            override_curations_by_id.get(idx, set())
                        ):
                            assert override_curation_to_modify.associated_id_sets is not None
                            new_associated_id_sets = self._drop_id_from_associated_id_sets(
                                idx, override_curation_to_modify.associated_id_sets
                            )
                            if len(new_associated_id_sets) == 0:

                                self.curations.remove(override_curation_to_modify)
                                override_curations_by_id[idx].remove(override_curation_to_modify)
                                logger.debug(
                                    "removed curation %s because of global action",
                                    override_curation_to_modify,
                                )
                            elif (
                                new_associated_id_sets
                                != override_curation_to_modify.associated_id_sets
                            ):
                                self.curations.remove(override_curation_to_modify)
                                override_curations_by_id[idx].remove(override_curation_to_modify)
                                mod_curation = dataclasses.replace(
                                    override_curation_to_modify,
                                    associated_id_sets=new_associated_id_sets,
                                )
                                self.curations.add(mod_curation)
                                override_curations_by_id[idx].add(mod_curation)
                                logger.debug(
                                    "modified curation %s to %s because of global action",
                                    override_curation_to_modify,
                                    mod_curation,
                                )

            else:
                raise ValueError(f"unknown behaviour for parser {self.parser_name}, {action}")
        return None

    def _attempt_to_add_database_entry_for_curation(
        self,
        curation_term_norm: NormalisedSynonymStr,
        curation_associated_id_set: Optional[AssociatedIdSets],
        curated_synonym: str,
    ) -> Literal[
        CurationModificationResult.SYNONYM_TERM_ADDED, CurationModificationResult.NO_ACTION
    ]:
        """Create a new :class:`~kazu.data.data.SynonymTerm` for the database, or return
        an existing matching one if already present.

        Notes:

        If a term_norm already exists in self._terms_by_term_norm that matches 'curation_term_norm',
        this method will check to see if the 'curation_associated_id_set' matches the existing terms
        :class:`.AssociatedIdSets`\\.

        If so, no action will be taken. If not, a warning will be logged as adding it
        will cause irregularities in the database.

        If the term_norm does not exist, this method will create a new :class:`~kazu.data.data.SynonymTerm`
        with the provided :class:`.AssociatedIdSets`\\.

        :param curation_term_norm:
        :param curation_associated_id_set:
        :param curated_synonym:
        :return:
        """
        log_prefix = "%(parser_name)s attempting to create synonym term for <%(synonym)s> term_norm: <%(term_norm)s> IDs: %(ids)s}"
        log_formatting_dict: dict[str, Any] = {
            "parser_name": self.parser_name,
            "synonym": curated_synonym,
            "term_norm": curation_term_norm,
            "ids": curation_associated_id_set,
        }

        # look up the term norm in the db
        maybe_existing_synonym_term = self._terms_by_term_norm.get(curation_term_norm)
        if curation_associated_id_set is None and maybe_existing_synonym_term is not None:
            logger.debug(
                log_prefix
                + " but no associated id set provided, so term will inherit the parser defaults",
                log_formatting_dict,
            )
            return CurationModificationResult.NO_ACTION
        elif curation_associated_id_set is None and maybe_existing_synonym_term is None:
            logger.error(
                log_prefix
                + " but term_norm <%(term_norm)s> does not exist in synonym database."
                + " Since no id set was provided, no entry can be created",
                log_formatting_dict,
            )
            return CurationModificationResult.NO_ACTION

        # curation_associated_id_set is implicitly not None
        assert curation_associated_id_set is not None
        if len(curation_associated_id_set) == 0:
            logger.debug(
                "all ids removed by global action for %s,  Parser name: %s",
                curated_synonym,
                self.parser_name,
            )
            return CurationModificationResult.NO_ACTION

        if maybe_existing_synonym_term is not None:
            log_formatting_dict["existing_id_set"] = maybe_existing_synonym_term.associated_id_sets

            if (
                len(
                    curation_associated_id_set.symmetric_difference(
                        maybe_existing_synonym_term.associated_id_sets
                    )
                )
                == 0
            ):
                logger.debug(
                    log_prefix
                    + " but term_norm <%(term_norm)s> already exists in synonym database."
                    + "since this SynonymTerm matches the id_set, no action is required. %(existing_id_set)s",
                    log_formatting_dict,
                )
                return CurationModificationResult.NO_ACTION
            else:
                logger.debug(
                    log_prefix
                    + " . Will remove existing term_norm <%(term_norm)s> as an ID set override has been specified",
                    log_formatting_dict,
                )

        # no term exists, or we want to override so one will be made
        assert curation_associated_id_set is not None
        for equiv_id_set in set(curation_associated_id_set):
            for idx in equiv_id_set.ids:
                if idx not in self._terms_by_id:
                    curation_associated_id_set = self._drop_id_from_associated_id_sets(
                        id_to_drop=idx, associated_id_sets=curation_associated_id_set
                    )
                    logger.warning(
                        "Attempted to add term containing %s but this id does not exist in the parser and will be ignored",
                        idx,
                    )
        if len(curation_associated_id_set) > 0:
            is_symbolic = StringNormalizer.classify_symbolic(curated_synonym, self.entity_class)
            new_term = SynonymTerm(
                term_norm=curation_term_norm,
                terms=frozenset((curated_synonym,)),
                is_symbolic=is_symbolic,
                mapping_types=frozenset(("kazu_curated",)),
                associated_id_sets=curation_associated_id_set,
                parser_name=self.parser_name,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            return self._update_term_lookups(new_term, True)
        else:
            return CurationModificationResult.NO_ACTION


class OntologyParser(ABC):
    """Parse an ontology (or similar) into a set of outputs suitable for NLP entity
    linking.

    Implementations should have a class attribute 'name' to
    something suitably representative. The key method is
    :meth:`~.OntologyParser.parse_to_dataframe`, which should convert an
    input source to a dataframe suitable for further processing.

    The other important method is :meth:`~.find_kb`. This should parse an ID
    string (if required) and return the underlying source. This is
    important for composite resources that contain identifiers from
    different seed sources.

    See :ref:`ontology_parser` for a more detailed guide.

    Generally speaking, when parsing a data source, synonyms that are
    symbolic (as determined by the :class:`~.StringNormalizer`) that refer to more
    than one id are more likely to be ambiguous. Therefore, we assume
    they refer to unique concepts (e.g. COX 1 could be 'ENSG00000095303'
    OR 'ENSG00000198804', and thus they will yield multiple instances of
    :class:`~.EquivalentIdSet`. Non symbolic synonyms (i.e. noun phrases) are far
    less likely to refer to distinct entities, so we might want to merge
    the associated ID's non-symbolic ambiguous synonyms into a single
    :class:`~.EquivalentIdSet`. The result of :meth:`.StringNormalizer.classify_symbolic`
    forms the ``is_symbolic`` parameter to :meth:`~.score_and_group_ids`.

    If the underlying knowledgebase contains more than one entity type,
    muliple parsers should be implemented, subsetting accordingly (e.g.
    MEDDRA_DISEASE, MEDDRA_DIAGNOSTIC).
    """

    # the synonym table should have these (and only these columns)
    all_synonym_column_names = [IDX, SYN, MAPPING_TYPE]
    # the metadata table should have at least these columns (note, IDX will become the index)
    minimum_metadata_column_names = [DEFAULT_LABEL, DATA_ORIGIN]

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        """
        :param in_path: Path to some resource that should be processed (e.g. owl file, db config, tsv etc)
        :param entity_class: The entity class to associate with this parser throughout the pipeline.
            Also used in the parser when calling StringNormalizer to determine the class-appropriate behaviour.
        :param name: A string to represent a parser in the overall pipeline. Should be globally unique
        :param string_scorer: Optional protocol of StringSimilarityScorer.  Used for resolving ambiguous symbolic
            synonyms via similarity calculation of the default label associated with the conflicted labels. If no
            instance is provided, all synonym conflicts will be assumed to refer to different concepts. This is not
            recommended!
        :param synonym_merge_threshold: similarity threshold to trigger a merge of conflicted synonyms into a single
            :class:`~.EquivalentIdSet`. See :meth:`~.OntologyParser.score_and_group_ids` for further details
        :param data_origin: The origin of this dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc. Note, this is different
            from the parser.name, as is used to identify the origin of a mapping back to a data source
        :param synonym_generator: optional CombinatorialSynonymGenerator. Used to generate synonyms for dictionary
            based NER matching
        :param curations_path: path to jsonl file of curations to apply to the parser
        :param global_actions:
        """

        self.in_path = in_path
        self.entity_class = entity_class
        self.name = name
        if string_scorer is None:
            logger.warning(
                "no string scorer configured for %s. Synonym resolution disabled.", self.name
            )
        self.string_scorer = string_scorer
        self.synonym_merge_threshold = synonym_merge_threshold
        self.data_origin = data_origin
        self.synonym_generator = synonym_generator
        self.curations_path = curations_path
        self.global_actions = global_actions
        self.parsed_dataframe: Optional[pd.DataFrame] = None
        self.metadata_db = MetadataDatabase()
        self.synonym_db = SynonymDatabase()

    @abstractmethod
    def find_kb(self, string: str) -> str:
        """Split an IDX somehow to find the ontology SOURCE reference.

        :param string: the IDX string to process
        :return:
        """
        pass

    def resolve_synonyms(self, synonym_df: pd.DataFrame) -> set[SynonymTerm]:

        result = set()
        synonym_df["syn_norm"] = synonym_df[SYN].apply(
            StringNormalizer.normalize, entity_class=self.entity_class
        )

        for i, row in (
            synonym_df[["syn_norm", SYN, IDX, MAPPING_TYPE]]
            .groupby(["syn_norm"])
            .agg(set)
            .reset_index()
            .iterrows()
        ):

            syn_set = row[SYN]
            mapping_type_set: frozenset[str] = frozenset(row[MAPPING_TYPE])
            syn_norm = row["syn_norm"]
            if len(syn_set) > 1:
                logger.debug("normaliser has merged %s into a single term: %s", syn_set, syn_norm)

            is_symbolic = all(
                StringNormalizer.classify_symbolic(x, self.entity_class) for x in syn_set
            )

            ids: set[str] = row[IDX]
            ids_and_source = set(
                (
                    idx,
                    self.find_kb(idx),
                )
                for idx in ids
            )
            associated_id_sets, agg_strategy = self.score_and_group_ids(ids_and_source, is_symbolic)

            synonym_term = SynonymTerm(
                term_norm=syn_norm,
                terms=frozenset(syn_set),
                is_symbolic=is_symbolic,
                mapping_types=mapping_type_set,
                associated_id_sets=associated_id_sets,
                parser_name=self.name,
                aggregated_by=agg_strategy,
            )

            result.add(synonym_term)

        return result

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
    ) -> tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """For a given data source, one normalised synonym may map to one or more id. In
        some cases, the ID may be duplicate/redundant (e.g. there are many chembl ids
        for paracetamol). In other cases, the ID may refer to distinct concepts (e.g.
        COX 1 could be 'ENSG00000095303' OR 'ENSG00000198804').

        Since synonyms from data sources are confused in such a manner, we need to decide some way to cluster them into
        a single :class:`~.SynonymTerm` concept, which in turn is a container for one or more :class:`~.EquivalentIdSet`
        (depending on whether the concept is ambiguous or not)

        The job of ``score_and_group_ids`` is to determine how many :class:`~.EquivalentIdSet`\\ s for a given set of
        ids should be produced.

        The default algorithm (which can be overridden by concrete parser implementations) works as follows:

        1. If no ``string_scorer`` is configured, create an :class:`~.EquivalentIdSet` for each id (strategy
           :attr:`~.EquivalentIdAggregationStrategy.NO_STRATEGY` - not recommended)
        2. If only one ID is referenced, or the associated normalised synonym string is not symbolic, group the
           ids into a single :class:`~.EquivalentIdSet` (strategy :attr:`~.EquivalentIdAggregationStrategy.UNAMBIGUOUS`)
        3. otherwise, compare the default label associated with each ID to every other default label. If it's above
           ``self.synonym_merge_threshold``, merge into one :class:`~.EquivalentIdSet`, if not, create a new one.

        recommendation: Use the :class:`~.SapbertStringSimilarityScorer` for comparison.

        .. important::
            Any calls to this method requires the metadata DB to be populated, as this is the store of
            :data:`~.DEFAULT_LABEL`.

        :param ids_and_source: ids to determine appropriate groupings of, and their associated sources
        :param is_symbolic: is the underlying synonym symbolic?
        :return:
        """
        if self.string_scorer is None:
            # the NO_STRATEGY aggregation strategy assumes all synonyms are ambiguous
            return (
                frozenset(
                    EquivalentIdSet(ids_and_source=frozenset((single_id_and_source,)))
                    for single_id_and_source in ids_and_source
                ),
                EquivalentIdAggregationStrategy.NO_STRATEGY,
            )
        else:

            if len(ids_and_source) == 1:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.UNAMBIGUOUS,
                )

            if not is_symbolic:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.MERGED_AS_NON_SYMBOLIC,
                )
            else:
                # use similarity to group ids into EquivalentIdSets

                DefaultLabels = set[str]
                id_list: list[tuple[IdsAndSource, DefaultLabels]] = []
                for id_and_source_tuple in ids_and_source:
                    default_label = cast(
                        str,
                        self.metadata_db.get_by_idx(self.name, id_and_source_tuple[0])[
                            DEFAULT_LABEL
                        ],
                    )
                    most_similar_id_set = None
                    best_score = 0.0
                    for id_and_default_label_set in id_list:
                        sim = max(
                            self.string_scorer(default_label, other_label)
                            for other_label in id_and_default_label_set[1]
                        )
                        if sim > self.synonym_merge_threshold and sim > best_score:
                            most_similar_id_set = id_and_default_label_set
                            best_score = sim

                    # for the first label, the above for loop is a no-op as id_sets is empty
                    # and the below if statement will be true.
                    # After that, it will be True if the id under consideration should not
                    # merge with any existing group and should get its own EquivalentIdSet
                    if not most_similar_id_set:
                        id_list.append(
                            (
                                {id_and_source_tuple},
                                {default_label},
                            )
                        )
                    else:
                        most_similar_id_set[0].add(id_and_source_tuple)
                        most_similar_id_set[1].add(default_label)

                return (
                    frozenset(
                        EquivalentIdSet(ids_and_source=frozenset(ids_and_source))
                        for ids_and_source, _ in id_list
                    ),
                    EquivalentIdAggregationStrategy.RESOLVED_BY_SIMILARITY,
                )

    def _parse_df_if_not_already_parsed(self):
        if self.parsed_dataframe is None:
            self.parsed_dataframe = self.parse_to_dataframe()
            self.parsed_dataframe[DATA_ORIGIN] = self.data_origin
            self.parsed_dataframe[IDX] = self.parsed_dataframe[IDX].astype(str)
            # since we always need a value for DEFAULT_LABEL,
            # if the underlying data doesn't provide one, just use the IDX
            rows_without_default_label = self.parsed_dataframe.loc[
                pd.isnull(self.parsed_dataframe[DEFAULT_LABEL])
            ]
            rows_without_default_label[DEFAULT_LABEL] = rows_without_default_label[IDX]

    @kazu_disk_cache.memoize(ignore={0})
    def export_metadata(self, parser_name: str) -> dict[str, dict[str, SimpleValue]]:
        """Export the metadata from the ontology.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return: {idx:{metadata_key:metadata_value}}
        """
        self._parse_df_if_not_already_parsed()
        assert self.parsed_dataframe is not None
        metadata_columns = self.parsed_dataframe.columns
        metadata_columns = metadata_columns.drop([MAPPING_TYPE, SYN])
        metadata_df = self.parsed_dataframe[metadata_columns]
        metadata_df = metadata_df.drop_duplicates(subset=[IDX])
        metadata_df = metadata_df.dropna(
            axis=0, subset=["idx"] + OntologyParser.minimum_metadata_column_names
        )
        metadata_df.set_index(inplace=True, drop=True, keys=IDX)
        assert set(OntologyParser.minimum_metadata_column_names).issubset(metadata_df.columns)
        metadata = metadata_df.to_dict(orient="index")
        return cast(dict[str, dict[str, SimpleValue]], metadata)

    def process_curations(
        self, terms: set[SynonymTerm]
    ) -> tuple[Optional[list[CuratedTerm]], set[SynonymTerm]]:
        if self.curations_path is None:
            logger.warning(
                "%s is configured to use raw ontology synonyms. This may result in noisy NER performance.",
                self.name,
            )
            curations_set = set()
            for term_str in extract_term_strings_from_synonym_terms(terms):
                curations_set.add(
                    string_to_putative_curation(
                        term_string=term_str,
                        entity_class=self.entity_class,
                    )
                )
            if self.synonym_generator is not None:
                logger.warning(
                    "%s is configured to use synonym generators. This may result in noisy NER performance.",
                    self.name,
                )
                curations_set.update(self.synonym_generator(curations_set))
            curations = list(curations_set)

        else:
            assert self.curations_path is not None
            logger.info(
                "%s is configured to use curations. Synonym generation will be ignored",
                self.name,
            )
            curations = load_curated_terms(self.curations_path)

        curation_processor = CurationProcessor(
            global_actions=self.global_actions,
            curations=curations,
            parser_name=self.name,
            entity_class=self.entity_class,
            synonym_terms=terms,
        )
        return curation_processor.export_curations_and_final_terms()

    @kazu_disk_cache.memoize(ignore={0})
    def export_synonym_terms(self, parser_name: str) -> set[SynonymTerm]:
        """Export :class:`.SynonymTerm` from the parser.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return:
        """
        self._parse_df_if_not_already_parsed()
        assert self.parsed_dataframe is not None
        # ensure correct order
        syn_df = self.parsed_dataframe[self.all_synonym_column_names].copy()
        syn_df = syn_df.dropna(subset=[SYN])
        syn_df[SYN] = syn_df[SYN].apply(str.strip)
        syn_df.drop_duplicates(subset=self.all_synonym_column_names)
        assert set(OntologyParser.all_synonym_column_names).issubset(syn_df.columns)
        synonym_terms = self.resolve_synonyms(synonym_df=syn_df)
        return synonym_terms

    @kazu_disk_cache.memoize(ignore={0})
    def _populate_databases(
        self, parser_name: str
    ) -> tuple[Optional[list[CuratedTerm]], dict[str, dict[str, SimpleValue]], set[SynonymTerm]]:
        """Disk cacheable method that populates all databases.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return:
        """
        logger.info("populating database for %s from source", self.name)
        metadata = self.export_metadata(self.name)
        # metadata db needs to be populated before call to export_synonym_terms
        self.metadata_db.add_parser(self.name, self.entity_class, metadata)
        intermediate_synonym_terms = self.export_synonym_terms(self.name)
        maybe_ner_curations, final_syn_terms = self.process_curations(intermediate_synonym_terms)
        self.parsed_dataframe = None  # clear the reference to save memory

        self.synonym_db.add(self.name, final_syn_terms)
        return maybe_ner_curations, metadata, final_syn_terms

    def populate_databases(
        self, force: bool = False, return_curations: bool = False
    ) -> Optional[list[CuratedTerm]]:
        """Populate the databases with the results of the parser.

        Also calculates the term norms associated with any curations (if provided) which
        can then be used for Dictionary based NER

        :param force: do not use the cache for the ontology parser
        :param return_curations: should processed curations be returned?
        :return: curations if required
        """
        if self.name in self.synonym_db.loaded_parsers and not force and not return_curations:
            logger.debug("parser %s already loaded.", self.name)
            return None

        cache_key = self._populate_databases.__cache_key__(self, self.name)

        if force:
            kazu_disk_cache.delete(self.export_metadata.__cache_key__(self, self.name))
            kazu_disk_cache.delete(self.export_synonym_terms.__cache_key__(self, self.name))
            kazu_disk_cache.delete(cache_key)

        maybe_curations, metadata, final_syn_terms = self._populate_databases(self.name)

        if self.name not in self.synonym_db.loaded_parsers:
            logger.info("populating database for %s from cache", self.name)
            self.metadata_db.add_parser(self.name, self.entity_class, metadata)
            self.synonym_db.add(self.name, final_syn_terms)

        return maybe_curations if return_curations else None

    @abstractmethod
    def parse_to_dataframe(self) -> pd.DataFrame:
        """Implementations should override this method, returning a 'long, thin' :class:`pandas.DataFrame` of at least the following
        columns:


        [:data:`~.IDX`, :data:`~.DEFAULT_LABEL`, :data:`~.SYN`, :data:`~.MAPPING_TYPE`]

        | :data:`~.IDX`: the ontology id
        | :data:`~.DEFAULT_LABEL`: the preferred label
        | :data:`~.SYN`: a synonym of the concept
        | :data:`~.MAPPING_TYPE`: the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology

        .. note:: It is the responsibility of the implementation of ``parse_to_dataframe`` to add default labels as synonyms.

        Any 'extra' columns will be added to the :class:`~kazu.database.in_memory_db.MetadataDatabase` as metadata fields for the
        given id in the relevant ontology.
        """
        pass



import logging
from collections.abc import Iterable, Callable
from functools import partial
from typing import Optional, cast

import torch
from torch import Tensor, sigmoid, softmax
from torch.utils.data import DataLoader
from transformers import (
    AutoModelForTokenClassification,
    AutoConfig,
    AutoTokenizer,
    DataCollatorWithPadding,
    BatchEncoding,
    PreTrainedTokenizerBase,
)
from transformers.file_utils import PaddingStrategy

from kazu.data.data import Document, Section
from kazu.data.pytorch import HFDataset
from kazu.steps import Step, document_batch_step
from kazu.steps.ner.entity_post_processing import NonContiguousEntitySplitter
from kazu.steps.ner.tokenized_word_processor import TokenizedWordProcessor, TokenizedWord
from kazu.utils.utils import documents_to_document_section_batch_encodings_map

logger = logging.getLogger(__name__)


class TransformersModelForTokenClassificationNerStep(Step):
    """A wrapper for :class:`transformers.AutoModelForTokenClassification`\\ .

    This implementation uses a sliding
    window concept to process large documents that don't fit into the maximum sequence length
    allowed by a model. Resulting token labels are then post processed by
    :class:`~kazu.steps.ner.tokenized_word_processor.TokenizedWordProcessor`.
    """

    def __init__(
        self,
        path: str,
        batch_size: int,
        stride: int,
        max_sequence_length: int,
        labels: list[str],
        detect_subspans: bool = False,
        threshold: Optional[float] = None,
        entity_splitter: Optional[NonContiguousEntitySplitter] = None,
        strip_re: Optional[dict[str, str]] = None,
    ):
        """

        :param path: path to HF model, config and tokenizer. Passed to HF .from_pretrained()
        :param batch_size: batch size for dataloader
        :param stride: passed to HF tokenizers (for splitting long docs)
        :param max_sequence_length: passed to HF tokenizers (for splitting long docs)
        :param labels:
        :param detect_subspans: attempt to detect nested entities (threshold must be configured)
        :param threshold: the confidence threshold used to detect nested entities
        :param entity_splitter: to detect non-contiguous entities if provided
        :param strip_re: passed to :class:`~kazu.steps.ner.tokenized_word_processor.TokenizedWordProcessor`
        """
        self.entity_splitter = entity_splitter
        if max_sequence_length % 2 != 0:
            raise RuntimeError(
                "max_sequence_length must %2 ==0 in order for correct document windowing"
            )
        self.max_sequence_length = max_sequence_length
        if stride % 2 != 0:
            raise RuntimeError("stride must %2 ==0 in order for correct document windowing")
        self.stride = stride
        self.batch_size = batch_size
        self.config = AutoConfig.from_pretrained(path)
        self.tokeniser: PreTrainedTokenizerBase = AutoTokenizer.from_pretrained(
            path, config=self.config
        )
        self.model = AutoModelForTokenClassification.from_pretrained(
            path, config=self.config
        ).eval()
        self.activation_fn = cast(
            Callable[[Tensor], Tensor], sigmoid if detect_subspans else partial(softmax, dim=-1)
        )
        self.tokenized_word_processor = TokenizedWordProcessor(
            detect_subspans=detect_subspans,
            confidence_threshold=threshold,
            id2label=self.id2labels_from_label_list(labels),
            strip_re=strip_re,
        )

    @document_batch_step
    def __call__(self, docs: list[Document]) -> None:
        loader, id_section_map = self.get_dataloader(docs)
        # need this so mypy knows to expect the dataset to have encodings
        dataset = cast(HFDataset, loader.dataset)
        # run the transformer and get results
        activations = self.get_activations(loader)
        for section_index, section in id_section_map.items():
            words = self.section_frames_to_tokenised_words(
                section_index=section_index,
                batch_encoding=dataset.encodings,
                predictions=activations,
            )
            entities = self.tokenized_word_processor(
                words, text=section.text, namespace=self.namespace()
            )
            section.entities.extend(entities)
            if self.entity_splitter:
                for ent in entities:
                    section.entities.extend(self.entity_splitter(ent, section.text))

    def get_activations(self, loader: DataLoader) -> Tensor:
        """Get a namedtuple_values_indices consisting of confidence and labels for a
        given dataloader (i.e. run bert)

        :param loader:
        :return:
        """
        with torch.no_grad():
            results = torch.cat(tuple(self.model(**batch).logits for batch in loader))
            return self.activation_fn(results)

    def get_dataloader(self, docs: list[Document]) -> tuple[DataLoader, dict[int, Section]]:
        """Get a dataloader from a List of :class:`kazu.data.data.Document`. Collation
        is handled via :class:`transformers.DataCollatorWithPadding`\\ .

        :param docs:
        :return: The returned dict's keys map to overflow_to_sample_mapping in the
            underlying batch encoding, allowing the processing of docs longer than can
            fit within the maximum sequence length of a transformer
        """
        batch_encoding, id_section_map = documents_to_document_section_batch_encodings_map(
            docs, self.tokeniser, stride=self.stride, max_length=self.max_sequence_length
        )
        dataset = HFDataset(batch_encoding)
        collate_func = DataCollatorWithPadding(
            tokenizer=self.tokeniser, padding=PaddingStrategy.MAX_LENGTH
        )
        loader = DataLoader(dataset=dataset, batch_size=self.batch_size, collate_fn=collate_func)
        return loader, id_section_map

    def frame_to_tok_word(
        self,
        batch_encoding: BatchEncoding,
        number_of_frames: int,
        frame_index: int,
        section_frame_index: int,
        predictions: Tensor,
    ) -> list[TokenizedWord]:
        """Depending on the number of frames generated by a string of text, and whether
        it is the first or last frame, we need to return different subsets of the frame
        offsets and frame word_ids.

        :param batch_encoding:
        :param number_of_frames: number of frames created by the tokenizer for the
            string
        :param frame_index: the index of the query frame, relative to the total number
            of frames
        :param section_frame_index: the index of the section frame, relative to the
            whole BatchEncoding
        :param predictions:
        :return: Tuple of 2 lists: frame offsets and frame word ids
        """
        half_stride = int(self.stride / 2)
        # 1:-1 skip cls and sep tokens
        if number_of_frames == 1:
            start_index = 1
            end_index = -1
        elif number_of_frames > 1 and frame_index == 0:
            start_index = 1
            end_index = -(half_stride + 1)
        elif number_of_frames > 1 and frame_index == number_of_frames - 1:
            start_index = half_stride + 1
            end_index = -1
        else:
            start_index = half_stride + 1
            end_index = -(half_stride + 1)

        assert batch_encoding.encodings is not None
        frame_offsets = batch_encoding.encodings[section_frame_index].offsets[start_index:end_index]
        frame_word_ids: list[Optional[int]] = batch_encoding.encodings[
            section_frame_index
        ].word_ids[start_index:end_index]
        frame_token_ids = batch_encoding.encodings[section_frame_index].ids[start_index:end_index]
        frame_tokens = batch_encoding.encodings[section_frame_index].tokens[start_index:end_index]
        predictions = predictions[section_frame_index][start_index:end_index]
        prev_word_id: Optional[int] = None
        all_words = []

        word_id_index_start, offset_start, offset_end = 0, 0, 0

        for i, word_id in enumerate(frame_word_ids):
            if word_id != prev_word_id:
                # start a new word, and add the previous
                if prev_word_id is not None:
                    all_words.append(
                        TokenizedWord(
                            token_offsets=frame_offsets[word_id_index_start:i],
                            token_confidences=predictions[word_id_index_start:i],
                            token_ids=frame_token_ids[word_id_index_start:i],
                            tokens=frame_tokens[word_id_index_start:i],
                            word_char_start=offset_start,
                            word_char_end=offset_end - 1,
                            word_id=prev_word_id,
                        )
                    )
                word_id_index_start = i
                offset_start, offset_end = frame_offsets[i]
            if i == len(frame_word_ids) - 1 and word_id is not None:
                # if checking the last word in a frame, if word_id is not None, add it
                all_words.append(
                    TokenizedWord(
                        token_offsets=frame_offsets[word_id_index_start : i + 1],
                        token_confidences=predictions[word_id_index_start : i + 1],
                        token_ids=frame_token_ids[word_id_index_start : i + 1],
                        tokens=frame_tokens[word_id_index_start : i + 1],
                        word_char_start=offset_start,
                        word_char_end=offset_end,
                        word_id=word_id,
                    )
                )

            _, offset_end = frame_offsets[i]
            prev_word_id = word_id

        frame_word_ids_excluding_nones = set(frame_word_ids)
        frame_word_ids_excluding_nones.discard(None)
        assert len(set(frame_word_ids_excluding_nones)) == len(all_words)

        logger.debug(
            f"inputs this frame: {self.tokeniser.decode(frame_token_ids[:start_index])}<-IGNORED "
            f"START->{self.tokeniser.decode(frame_token_ids[start_index:end_index])}"
            f"<-END IGNORED->{self.tokeniser.decode(frame_token_ids[end_index:])}"
        )
        return all_words

    def section_frames_to_tokenised_words(
        self,
        section_index: int,
        batch_encoding: BatchEncoding,
        predictions: Tensor,
    ) -> list[TokenizedWord]:
        """

        :param section_index:
        :param batch_encoding:
        :param predictions:
        :return:
        """
        words = []

        section_frame_indices = self.get_list_of_batch_encoding_frames_for_section(
            batch_encoding, section_index
        )
        for frame_index, section_frame_index in enumerate(section_frame_indices):
            word_sub_list = self.frame_to_tok_word(
                batch_encoding=batch_encoding,
                number_of_frames=len(section_frame_indices),
                frame_index=frame_index,
                section_frame_index=section_frame_index,
                predictions=predictions,
            )
            words.extend(word_sub_list)
        return words

    @staticmethod
    def get_list_of_batch_encoding_frames_for_section(
        batch_encoding: BatchEncoding, section_index: int
    ) -> list[int]:
        """For a given dataloader with a HFDataset, return a list of frame indexes
        associated with a given section index.

        :param batch_encoding:
        :param section_index:
        :return:
        """
        section_frame_indices = [
            i
            for i, x in enumerate(batch_encoding.data["overflow_to_sample_mapping"])
            if x == section_index
        ]
        return section_frame_indices

    @staticmethod
    def id2labels_from_label_list(labels: Iterable[str]) -> dict[int, str]:
        return {idx: label for idx, label in enumerate(labels)}



import argparse
import dataclasses
import os
from collections import defaultdict
from pathlib import Path

from hydra import initialize_config_dir, compose
from hydra.utils import instantiate
from omegaconf import open_dict

from kazu.data.data import CuratedTerm
from kazu.utils.constants import HYDRA_VERSION_BASE

_SOURCE_TERMS_PREFIX = "source_terms"
_GENERATED_TERMS_PREFIX = "generated_terms"
_MIGRATED_TERMS_FILENAME = "_migrated_terms.jsonl"
_EXTRA_ONTOLOGY_TERMS_FILENAME = "_extra_ontology_terms.jsonl"
_MODIFIED_TERMS_FILENAME = "_modified_terms.jsonl"
_CASE_WARNING_TERMS_FILENAME = "_matched_terms_case_warnings.jsonl"
_OBSOLETE_TERMS_FILENAME = "_obsolete_terms.jsonl"
_NOVEL_TERMS_FILENAME = "_novel_terms.jsonl"

CURATION_REPORT_INSTRUCTIONS = f"""The subdirectories contain various jsonl files of how the current set of CuratedTerms map to the configured ontology.
{_MIGRATED_TERMS_FILENAME}: terms that are compatible with the configured version of the ontology
{_MODIFIED_TERMS_FILENAME}: terms that match but only if case is not considered. Since only one set of behaviours exist between the original CuratedTerm and the new match, these CuratedTerms have been migrated to the new match.
{_CASE_WARNING_TERMS_FILENAME}: terms that match if case is not considered, but with multiple candidates in the new Ontology. These need to be recurated.
{_OBSOLETE_TERMS_FILENAME}: terms from the set of CuratedTerms that are now obsolete and can be removed.
{_NOVEL_TERMS_FILENAME}: terms from the Ontology that are novel and need to be curated.
{_EXTRA_ONTOLOGY_TERMS_FILENAME}: manually added terms that don't exist in the ontology, or are mapped to different id's than the ontology default

To summarise what to do:

recurate the following files:
{_SOURCE_TERMS_PREFIX}{_CASE_WARNING_TERMS_FILENAME}
{_SOURCE_TERMS_PREFIX}{_NOVEL_TERMS_FILENAME}
{_GENERATED_TERMS_PREFIX}{_CASE_WARNING_TERMS_FILENAME}
{_GENERATED_TERMS_PREFIX}{_NOVEL_TERMS_FILENAME}

then run

rm  *{_OBSOLETE_TERMS_FILENAME}

cat *.jsonl > """


class _OntologyUpgradeReport:
    def __init__(
        self,
        existing_curations: set[CuratedTerm],
        curations_from_new_ontology_version: set[CuratedTerm],
    ):

        self.incoming_curations_case_sensitive: defaultdict[str, set[CuratedTerm]] = defaultdict(
            set
        )
        self.incoming_curations_case_insensitive: defaultdict[str, set[CuratedTerm]] = defaultdict(
            set
        )
        for curation in curations_from_new_ontology_version:
            self.incoming_curations_case_sensitive[curation.curated_synonym].add(curation)
            self.incoming_curations_case_insensitive[curation.curated_synonym.lower()].add(curation)

        self.existing_curations_case_sensitive: defaultdict[str, set[CuratedTerm]] = defaultdict(
            set
        )
        for curation in existing_curations:
            self.existing_curations_case_sensitive[curation.curated_synonym].add(curation)

        # old/new curations that perfectly match on .curated_synonym
        self.matched_curations: set[CuratedTerm] = set()

        # a pointer to new curations that can be discarded
        self.eliminated_generated_curations_case_insensitive: defaultdict[
            str, set[CuratedTerm]
        ] = defaultdict(set)

        # curations that match in a case insensitive fashion, and can be automatically modified
        self.modified_curations: set[CuratedTerm] = set()

        # terms that don't appear in the ontology, but have been manually added
        self.extra_ontology_terms: set[CuratedTerm] = set()

        # terms that match only if case not considered, but with multiple candidates in the new Ontology.
        self.failed_migrations: defaultdict[str, set[CuratedTerm]] = defaultdict(set)

        # terms that no longer appear in the ontology
        self.obsolete_terms: defaultdict[str, set[CuratedTerm]] = defaultdict(set)

        # terms that are new in the latest version of the ontology
        self.novel_terms: defaultdict[str, set[CuratedTerm]] = defaultdict(set)

    def match_existing_terms_to_new_terms(
        self,
    ) -> None:

        for existing_term_string_case_sensitive, existing_terms in list(
            self.existing_curations_case_sensitive.items()
        ):
            matched_generated_terms = self.incoming_curations_case_sensitive.pop(
                existing_term_string_case_sensitive, None
            )
            if matched_generated_terms is not None:
                # we were able to find a match, so this set of terms is still good
                self.matched_curations.update(existing_terms)
                # pop this key as it's already been handled and we don't need to worry about it any more
                self.existing_curations_case_sensitive.pop(existing_term_string_case_sensitive)
                # keep track of the generated terms for this string, so we know that the case sensitive versions have been
                # handled. We'll need this later when we look at case insensitive matching
                self.eliminated_generated_curations_case_insensitive[
                    existing_term_string_case_sensitive.lower()
                ].update(matched_generated_terms)

        # now loop again over the remaining items, considering the more difficult cases where exact matching wasn't possible
        # we need to loop twice as we want to ensure every case sensitive match first, to ensure self.eliminated_generated_curations_ci
        # is fully populated and we don't miss any curations that could be mapped.
        for (
            existing_term_string_case_sensitive,
            existing_terms,
        ) in self.existing_curations_case_sensitive.items():
            matched_generated_terms_ci = self.incoming_curations_case_insensitive.pop(
                existing_term_string_case_sensitive.lower(), None
            )
            if matched_generated_terms_ci is None:
                # if no match, term is either extra ontological (i.e. added separately) or obsolete
                for term in existing_terms:
                    if term.additional_to_source:
                        self.extra_ontology_terms.add(term)
                    elif (
                        term.source_term is None
                    ):  # we group by source term for easy reviewing later
                        self.obsolete_terms[term.curated_synonym].add(term)
                    else:
                        self.obsolete_terms[term.source_term].add(term)
            else:
                # if there's a case insensitive match, remove any curations that have already been handled by case sensitive matching
                matched_generated_terms_ci_minus_already_handled_ones = (
                    matched_generated_terms_ci.difference(
                        self.eliminated_generated_curations_case_insensitive[
                            existing_term_string_case_sensitive.lower()
                        ]
                    )
                )
                if len(matched_generated_terms_ci_minus_already_handled_ones) == 0:
                    continue
                # we can only migrate a term if there is only a single set of behaviours from the original curation
                human_behaviour_set = set(x.control_aspects for x in existing_terms)
                if len(human_behaviour_set) == 1:
                    behaviour, case_sensitive, associated_id_sets = next(iter(human_behaviour_set))
                    for generated_term in matched_generated_terms_ci_minus_already_handled_ones:
                        self.modified_curations.add(
                            dataclasses.replace(
                                generated_term,
                                associated_id_sets=associated_id_sets,
                                behaviour=behaviour,
                                case_sensitive=case_sensitive,
                            )
                        )
                else:
                    # add to failure list, so the generated terms can be curated
                    for generated_term in matched_generated_terms_ci_minus_already_handled_ones:
                        self.failed_migrations[generated_term.curated_synonym].add(generated_term)

        # any remaining terms are novel
        for (
            generated_term_string_case_insensitive,
            matched_generated_terms_ci,
        ) in self.incoming_curations_case_insensitive.items():
            matched_generated_terms_ci_minus_already_handled_ones = (
                matched_generated_terms_ci.difference(
                    self.eliminated_generated_curations_case_insensitive[
                        generated_term_string_case_insensitive.lower()
                    ]
                )
            )
            for term in matched_generated_terms_ci_minus_already_handled_ones:
                if term.source_term is None:
                    self.novel_terms[term.curated_synonym].add(term)
                else:
                    self.novel_terms[term.source_term].add(term)

    @staticmethod
    def _write_unsorted_curation_set(path: Path, curation_set: set[CuratedTerm]) -> None:
        with path.open(mode="w") as f:
            for curation in curation_set:
                f.write(f"{curation.to_json()}\n")

    @staticmethod
    def _write_curations_shortest_synonyms_first(
        path: Path, curations: dict[str, set[CuratedTerm]]
    ) -> None:
        with path.open(mode="w") as f:
            by_len = sorted(curations.items(), key=lambda x: len(x[0]), reverse=False)
            for _term_str, terms in by_len:
                for curation in terms:
                    f.write(f"{curation.to_json()}\n")

    @classmethod
    def build_match_and_write_report(
        cls,
        existing_curations: set[CuratedTerm],
        curations_from_new_ontology_version: set[CuratedTerm],
        output_path: Path,
        parser_name: str,
        prefix: str,
        curation_file_name: str,
    ) -> None:
        report = cls(
            existing_curations=existing_curations,
            curations_from_new_ontology_version=curations_from_new_ontology_version,
        )
        report.match_existing_terms_to_new_terms()
        report.write_curation_report(
            output_path=output_path,
            parser_name=parser_name,
            prefix=prefix,
            curation_file_name=curation_file_name,
        )

    def write_curation_report(
        self, output_path: Path, parser_name: str, prefix: str, curation_file_name: str
    ) -> None:
        results_dir = output_path.joinpath("curation_reports").joinpath(parser_name)
        results_dir.mkdir(exist_ok=True, parents=True)
        self._write_unsorted_curation_set(
            results_dir.joinpath(f"{prefix}{_MIGRATED_TERMS_FILENAME}"), self.matched_curations
        )
        self._write_unsorted_curation_set(
            results_dir.joinpath(f"{prefix}{_EXTRA_ONTOLOGY_TERMS_FILENAME}"),
            self.extra_ontology_terms,
        )
        self._write_unsorted_curation_set(
            results_dir.joinpath(f"{prefix}{_MODIFIED_TERMS_FILENAME}"), self.modified_curations
        )

        self._write_curations_shortest_synonyms_first(
            results_dir.joinpath(f"{prefix}{_CASE_WARNING_TERMS_FILENAME}"), self.failed_migrations
        )
        self._write_curations_shortest_synonyms_first(
            results_dir.joinpath(f"{prefix}{_OBSOLETE_TERMS_FILENAME}"), self.obsolete_terms
        )
        self._write_curations_shortest_synonyms_first(
            results_dir.joinpath(f"{prefix}{_NOVEL_TERMS_FILENAME}"), self.novel_terms
        )
        with results_dir.joinpath("instructions.txt").open(mode="w") as f:
            f.write(CURATION_REPORT_INSTRUCTIONS + curation_file_name)


def run_curation_report(model_pack_path: Path) -> None:
    """Assists with upgrading CuratedTerms to a new version of an Ontology.

    .. warning::

       Calling this function will invalidate the kazu disk cache in the model pack.

    Since the currency of NER is strings, but the currency of entity
    linking is identifiers, it can be challenging to migrate an existing
    set of CuratedTerms to a new version of an ontology. Many things can
    change, such as terms become obsolete, their case may change, and/or
    new terms can be introduced.

    All the while, one wants to be able to reuse as many of the human
    curations one has made as possible.

    The aim of this class is to assist with this process as far as is
    possible, by producing a report that migrates previous curations,
    modifies existing curations to match their equivalents in a new
    version of an ontology, and highlights obsolete/new terms for
    curation.

    The algorithm runs as follows:

    1. Run synonym generation routines on the new version of the ontology to produce a set of :class:`.CuratedTerm`\\.

    2. Match generated synonyms with existing curations on a case-sensitive basis.

    3. Per synonym, if a match is detected, keep the original :class:`.CuratedTerm` and disregard any generated
       :class:`.CuratedTerm`\\s.

    4. For any remaining existing curations, match again in a case-insensitive fashion.

    5. If no match is detected, the term is either manually added, or is now obsolete in the original ontology.

    6. If a match is detected, the term may be migrated. This means copying the behaviour of the original
       :class:`.CuratedTerm` to the newly generated term. This is only possible if a single set of control aspects
       is defined for this synonym (defined by :attr:`.CuratedTerm.control_aspects`\\. Otherwise, the term
       will have to be manually curated again.

    7. If no match is detected, the newly generated :class:`.CuratedTerm` is novel to this version of the ontology and
       must be curated.

    A curation report directory will be created in the model pack, with subdirectories per parser. Follow on instructions
    can be found in these directories.
    """

    # relative imports so function can be imported without initialising disk cache
    from kazu.utils.caching import kazu_disk_cache
    from kazu.ontology_preprocessing.base import load_curated_terms

    with initialize_config_dir(
        version_base=HYDRA_VERSION_BASE, config_dir=str(model_pack_path.joinpath("conf"))
    ):
        cfg = compose(
            config_name="config",
            overrides=[
                "hydra/job_logging=none",
                "hydra/hydra_logging=none",
            ],
        )
        for parser_name, parser_cfg in cfg.ontologies.parsers.items():
            curations_path = Path(parser_cfg.curations_path)

            existing_source_term_curations = set()
            existing_generated_curations = set()
            for human_curated_term in load_curated_terms(curations_path):
                if human_curated_term.source_term is None:
                    existing_source_term_curations.add(human_curated_term)
                else:
                    existing_generated_curations.add(human_curated_term)

            new_source_term_curations = set()
            new_generated_curations = set()
            with open_dict(parser_cfg):
                parser_cfg.curations_path = None
                parser = instantiate(parser_cfg, _convert_="all")
                for generated_curated_term in parser.populate_databases(
                    return_curations=True, force=True
                ):
                    if generated_curated_term.source_term is None:
                        new_source_term_curations.add(generated_curated_term)
                    else:
                        new_generated_curations.add(generated_curated_term)

            # original terms report
            _OntologyUpgradeReport.build_match_and_write_report(
                existing_curations=existing_source_term_curations,
                curations_from_new_ontology_version=new_source_term_curations,
                output_path=model_pack_path,
                parser_name=parser_name,
                prefix=_SOURCE_TERMS_PREFIX,
                curation_file_name=curations_path.name,
            )

            # generated terms report
            _OntologyUpgradeReport.build_match_and_write_report(
                existing_curations=existing_generated_curations,
                curations_from_new_ontology_version=new_generated_curations,
                output_path=model_pack_path,
                parser_name=parser_name,
                prefix=_GENERATED_TERMS_PREFIX,
                curation_file_name=curations_path.name,
            )
            kazu_disk_cache.clear()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="write a curation report to explore and update curated terms between different ontology versions."
    )

    parser.add_argument(
        "--model_pack_path",
        type=Path,
        required=True,
        help="""Path to the model pack that contains the updated ontology(s) and the original curations.""",
    )
    path: Path = parser.parse_args().model_pack_path
    print(f"setting KAZU_MODEL_PACK to {path}")
    os.environ["KAZU_MODEL_PACK"] = str(path.absolute())
    run_curation_report(path)



import logging
from typing import TypedDict, Protocol

import torch
from tokenizers import Encoding
from torch import Tensor
from torch.utils.data import DataLoader, Dataset
from transformers import (
    AutoConfig,
    AutoTokenizer,
    AutoModel,
    BatchEncoding,
    DataCollatorWithPadding,
)
from transformers.file_utils import PaddingStrategy

from kazu.utils.utils import Singleton

logger = logging.getLogger(__name__)


class _BatchEncodingData(TypedDict):
    """Necessary because transformers doesn't distribute type information.

    We've only declared the properties we need currently, there may be more.
    """

    input_ids: dict[int, Tensor]
    token_type_ids: dict[int, Tensor]
    attention_mask: dict[int, Tensor]
    indices: dict[int, Tensor]


class _BatchEncodingFastTokenized(Protocol):
    """Necessary because transformers doesn't distribute type information.

    Also, we want to be able to say we know this was encoded with a fast tokenizer, so
    encodings won't be None.
    """

    @property
    def encodings(self) -> list[Encoding]:
        raise NotImplementedError

    @property
    def data(self) -> _BatchEncodingData:
        raise NotImplementedError


class HFSapbertInferenceDataset(Dataset[dict[str, Tensor]]):
    """A dataset to be used for inferencing.

    In addition to standard BERT  encodings, this uses an 'indices' encoding that can be
    used to track the vector index of an embedding. This is needed in a multi GPU
    environment.
    """

    def __init__(self, encodings: BatchEncoding):
        """Simple implementation of IterableDataset, producing HF tokenizer input_id.

        :param encodings: Expected to be produced by a :class:`transformers.PreTrainedTokenizerFast`.
            'slow' tokenizers (:class:`transformers.PreTrainedTokenizer`) store their encodings
            differently and so won't work with this class as-is.
        """
        if not encodings.is_fast:
            raise TypeError(
                "Your encodings come from a 'slow' tokenizer, only 'fast' tokenizers are supported."
            )
        self.encodings: _BatchEncodingFastTokenized = encodings

    def __len__(self) -> int:
        return len(self.encodings.encodings)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        query_toks1 = {
            "input_ids": self.encodings.data["input_ids"][index],
            "token_type_ids": self.encodings.data["token_type_ids"][index],
            "attention_mask": self.encodings.data["attention_mask"][index],
            "indices": self.encodings.data["indices"][index],
        }
        return query_toks1


class SapBertHelper(metaclass=Singleton):
    """Helper class to wrap useful SapBert inference functions.

    Original source:

    https://github.com/cambridgeltl/sapbert

    Licensed under MIT

    Copyright (c) Facebook, Inc. and its affiliates.

    .. raw:: html

        <details>
        <summary>Full License</summary>

    MIT License

    Copyright (c) Facebook, Inc. and its affiliates.

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    .. raw:: html

        </details>

    Paper:

    | Fangyu Liu, Ehsan Shareghi, Zaiqiao Meng, Marco Basaldella, and Nigel Collier. 2021.
    | `Self-alignment pretraining for biomedical entity representations. <https://www.aclweb.org/anthology/2021.naacl-main.334.pdf>`_
    | In Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies, pages 4228–4238.

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @inproceedings{liu2021self,
            title={Self-Alignment Pretraining for Biomedical Entity Representations},
            author={Liu, Fangyu and Shareghi, Ehsan and Meng, Zaiqiao and Basaldella, Marco and Collier, Nigel},
            booktitle={Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies},
            pages={4228--4238},
            month = jun,
            year={2021}
        }

    .. raw:: html

        </details>
    """

    def __init__(self, path: str):
        """

        :param path: passed to :class:`transformers.AutoConfig`\\, :class:`transformers.AutoTokenizer`\\, :class:`transformers.AutoModel` .from_pretrained.
        """
        self.config = AutoConfig.from_pretrained(path)
        self.tokenizer = AutoTokenizer.from_pretrained(path, config=self.config)
        self.model = AutoModel.from_pretrained(path, config=self.config)

    @staticmethod
    def get_embeddings(output: list[dict[int, torch.Tensor]]) -> torch.Tensor:
        """Get a tensor of embeddings in original order.

        :param output: int is the original index of the input.
        :return:
        """
        full_dict = {}
        for batch in output:
            full_dict.update(batch)
        if len(full_dict) > 1:
            embedding = torch.squeeze(torch.cat(list(full_dict.values())))
        else:
            embedding = torch.cat(list(full_dict.values()))
        return embedding

    def get_embeddings_from_dataloader(self, loader: DataLoader[BatchEncoding]) -> torch.Tensor:
        """Get the cls token output from all data in a dataloader as a 2d tensor.

        :param loader:
        :return: 2d tensor of cls  output
        """
        model = self.model.eval()
        predictions = []
        with torch.no_grad():
            for batch in loader:
                predictions.append(self.get_prediction_from_batch(model, batch))
        results = self.get_embeddings(predictions)
        return results

    @staticmethod
    def get_prediction_from_batch(
        model: torch.nn.Module, batch: dict[str, Tensor]
    ) -> dict[int, Tensor]:
        """Pass the batch through the model, and return the Tensor of the cls token."""

        indices = batch.pop("indices")
        batch_embeddings = model(**batch)
        cls_tokens = batch_embeddings.last_hidden_state[:, 0, :]
        # put index as dict key so we can realign the embedding space
        return {
            index.item(): cls_tokens[[batch_index], :] for batch_index, index in enumerate(indices)
        }

    def get_embedding_dataloader_from_strings(
        self,
        texts: list[str],
        batch_size: int,
        num_workers: int,
        max_length: int = 50,
    ) -> DataLoader[BatchEncoding]:
        """Get a dataloader with dataset :class:`.HFSapbertInferenceDataset` and
        DataCollatorWithPadding.

        This should be used to generate embeddings for strings of interest.

        :param texts: strings to use in the dataset
        :param batch_size:
        :param num_workers:
        :param max_length:
        :return:
        """
        indices = [i for i in range(len(texts))]
        # padding handled by collate func
        batch_encodings = self.tokenizer(
            texts, padding=PaddingStrategy.MAX_LENGTH, max_length=max_length, truncation=True
        )
        batch_encodings["indices"] = indices
        dataset = HFSapbertInferenceDataset(batch_encodings)
        loader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            collate_fn=DataCollatorWithPadding(
                tokenizer=self.tokenizer, padding=PaddingStrategy.LONGEST
            ),
            num_workers=num_workers,
        )
        return loader

    def get_embeddings_for_strings(self, texts: list[str], batch_size: int = 16) -> torch.Tensor:
        """For a list of strings, generate embeddings.

        This is a convenience function for users, as we need to carry out these steps
        several times in the codebase.

        :param texts:
        :param batch_size: optional batch size to use. If not specified, use 16
        :return: a 2d tensor of embeddings
        """

        loader = self.get_embedding_dataloader_from_strings(texts, batch_size, 0)
        results = self.get_embeddings_from_dataloader(loader)
        return results
